#include <stdio.h>
#include <stdlib.h>

typedef unsigned char uchar;

typedef enum pin_state_ {
  PIN_STATE_INIT,
  PIN_STATE_ON,
  PIN_STATE_FADE,
  PIN_STATE_OFF
} pin_state_t;
  
typedef struct pin_info_ {
  pin_state_t state;
  uchar       pwm_counter;
  uchar       pwm;
  uchar       level;
  uchar       time1;
  uchar       timex;
  unsigned    total_on;

  uchar       t_on;
  uchar       t_fade;
  uchar       t_off;
} pin_info_t;

static pin_info_t pins[7];
static uchar pwm_period_counter;

static void pwm_driver(void)
{
  size_t ii;

  pwm_period_counter++;
  for (ii = 0; ii < 7; ii++) {
    pins[ii].pwm_counter--;
  }
  if (0 == pwm_period_counter) {
    for (ii = 0; ii < 7; ii++) {
      pins[ii].pwm_counter++;
    }
  }
}

static unsigned short seed;
static uchar prand(void)
{
  seed = (seed + 7919) ^ (seed << 2);
  return (seed >> 5) & 0x1f;
}

/* 2 <= t_on <= 3    */
/* 3 <= t_fade <= 9  */
/* 30 <= t_off <= 76 */

static struct {
  unsigned t_on[2];
  unsigned t_fade[8];
  unsigned t_off[32];
} stats;

static void fader(
    pin_info_t *pin, 
    uchar t_on, 
    uchar t_fade, 
    uchar t_off)
{
  if (PIN_STATE_INIT == pin->state) {
    pin->level = 255;
    pin->pwm   = pin->level;
    pin->time1 = 2 + (prand() & 1); /* t_on; */
    pin->t_on  = pin->time1;
    stats.t_on[pin->time1 - 2]++;
    pin->timex = 0;
    pin->state++;
  }
  if (PIN_STATE_ON == pin->state) {
    pin->total_on += 3;
    pin->timex++;
    if (pin->timex >= 50) {
      pin->timex = 0;
      pin->time1--;
      if (0 == pin->time1) {
        pin->state++;
        pin->timex = 3 + (prand() & 7); /* t_fade; */
        stats.t_fade[pin->timex - 3]++;
        pin->t_fade = pin->timex;
      }
    }
  }
  if (PIN_STATE_FADE == pin->state) {
    pin->total_on += 3;
    if (pin->timex >= 5) {
      pin->timex -= 5;
    } else {
      uchar tmp;

      tmp = 3 + (prand() & 7);
#if 0
      /*if (pin->t_fade != tmp) */{
        printf("fade was(%u) is (%u)\n", pin->t_fade, tmp);
      }
#endif
      pin->timex = tmp; /* t_fade; */
      pin->level--;
      if (pin->level == 0) {
        pin->time1 = 30 + (prand() & 0x1f); /* t_off; */
        stats.t_off[pin->time1-30]++;
        pin->t_off = pin->time1;
        pin->timex = 0;
        pin->state++;
      }
    }
  }
  if (PIN_STATE_OFF == pin->state) {
    pin->timex++;
    if (pin->timex == 250) {
      pin->timex = 0;
      pin->time1--;
      if (0 == pin->time1) {
        static unsigned count = 0;

        pin->state = PIN_STATE_INIT;
        if (count++ < 20) {
          printf("t_on(%u) t_fade(%u) t_off(%u)\n",
              pin->t_on, pin->t_fade, pin->t_off);
        }
      }
    }
  }
}

int main(int argc, char **argv)
{
  size_t ii;

  for (ii = 0; ii < 7; ii++) {
    pins[ii].state = 0;
    pins[ii].total_on = 0;
  }
  for (ii = 0; ii < 3600000; ii += 3) {
    fader(pins + 0, 2, 6, 2 * 15);
    pwm_driver();
    fader(pins + 1, 3, 7, 2 * 19);
    pwm_driver();
    fader(pins + 2, 2, 9, 2 * 22);
    pwm_driver();
    fader(pins + 3, 3, 5, 2 * 25);
    pwm_driver();
    fader(pins + 4, 2, 6, 2 * 29);
    pwm_driver();
    fader(pins + 5, 3, 3, 2 * 35);
    pwm_driver();
    fader(pins + 6, 2, 4, 2 * 38);
  }

  for (ii = 0; ii < 7; ii++) {
    printf("total on(%u): %u\n",
        ii, pins[ii].total_on);
  }
  printf("t_on:");
  for (ii = 0; ii < 2; ii++) {
    printf(" %u", stats.t_on[ii]);
  }
  printf("\nt_fade:");
  for (ii = 0; ii < 8; ii++) {
    printf(" %u", stats.t_fade[ii]);
  }
  printf("\nt_off:");
  for (ii = 0; ii < 32; ii++) {
    printf(" %u", stats.t_off[ii]);
  }
  printf("\n");
}

